<?php
/**
 * generator.php
 *
 * This code was written for the CCGHQ community and is public domain.
 * Do whatever you want with this!
 *
 * Version 1 - 2nd/aug/2016
 *   First script that generates a full image, will be updated with the data file
 *   feeding script once I figure out our to interpret it.
 *
 */

/**
 * Token puzzle renderer class
 */
class TokenPuzzleRenderer {
  /**
   * Relative or absolute path to the assets directory
   *
   * @var array
   */
  private static $AssetsPath = "Rendering Assets";

  /**
   * Relative or absolute path to the data file
   *
   * @var string
   */
  private static $DataFile = "Puzzle Text Eternal Masters.csv";

  /**
   * Assets file names
   *
   * @var array
   */
  private static $Assets = array(
    'background' => 'framePost8thBlack.png',
    'layout-1'   => 'Chrome Mox.jpg',
    'layout-2'   => 'Firebolt.jpg',
    'layout-3'   => 'Force of Will.jpg',
    'layout-4'   => 'Hydroblast.jpg',
    'layout-5'   => 'Jace, the Mind Sculptor.jpg',
    'layout-6'   => 'Memory Lapse.jpg',
    'layout-7'   => 'Timberwatch Elf.jpg',
    'symbol-1'   => 'rose-final-resized.png',
    'symbol-2'   => 'crown-final-resized.png',
    'font'       => 'PlantinMTPro-Regular.otf');

  /**
   * Internal static data
   *
   * @var array
   */
  private static $Data = array(
    'layout-offset-top'  => 70,
    'layout-offset-left' => 70,
    'font-size'          => 40,
    'text-offset-top'    => 18,
    'text-offset-right'  => 22,
    'text-offset-bottom' => 18,
    'text-offset-left'   => 22,
    'text-color'         => array(255, 255, 255));

  /**
   * Internal cache for layout image resources
   *
   * @var array
   */
  private $_layouts_resources = array();

  /**
   * Internal image being generated
   *
   * @var resource
   */
  private $_image = null;

  /**
   * Width of the image being generated in pixels
   *
   * @var int
   */
  private $_width = null;

  /**
   * Height of the image being generated in pixels
   *
   * @var int
   */ 
  private $_height = null;

  /**
   * Internal image color cache - Text color
   *
   * @var resource
   */
  private $_textColor = null;

  /**
   * Default constructor
   */
  public function __construct() {
    if (!is_dir(self::$AssetsPath))
      throw new Exception("Cannot locate assets directory \"" .
            self::$AssetsPath . "\" (current directory: \"" .
            getcwd() . "\")");
  }

  /**
   * Retrieve asset path
   *
   * @param string $key Asset internal name
   * @return string Path to the asset file
   */
  private function _getAsset($key) {
    if (!isset(self::$Assets[$key]))
      throw new Exception("Invalid undefined asset \"$key\"");

    $path = self::$AssetsPath . DIRECTORY_SEPARATOR . self::$Assets[$key];
    if (!file_exists($path))
      throw new Exception("Missing asset \"$key\" in \"$path\"");

    return $path;
  }

  /**
   * Write text in a predefined position with trailing symbol support
   *
   * The symbol can only be at the end of the text, and referenced using
   * the special tag "[1]" or "[2]", the index number must match the asset
   * internal name "symbol-1", "symbol-2" and so on.
   *
   * The position index can be one of:
   *  - 0   top text
   *  - 1   right text
   *  - 2   bottom text
   *  - 3   left text
   *
   * @param int $pos_index Position index (top, right, bottom, left)
   * @param string $text Text to write
   */
  private function _writeText($pos_index, $text) {
    /* text rotation */
    $angle = ($pos_index % 2) * 90;

    /* extract symbol markers */
    if (preg_match('/\[([0-9]+)\]$/', $text, $regp)) {
      $symbol_file = $this->_getAsset("symbol-" . $regp[1]);
      /* we replace the symbol marker with spaces to keep the bounding box
       * space, required for horizontal and vertical centering */
      $text = str_replace($regp[0], "        ", $text);

      $im_symbol = imagecreatefrompng($symbol_file);
      if ($angle) {
        $tmp = imagerotate($im_symbol, $angle, 0);
        imagedestroy($im_symbol);
        $im_symbol = $tmp;
        unset($tmp);
      }

      $_symbol_width = imagesx($im_symbol);
      $_symbol_height = imagesy($im_symbol);
    }
    else { // no symbols this time
      $symbol_file = null;
      $im_symbol = null;
      $_symbol_width = 0;
      $_symbol_height = 0;
    }

    /*   (6, 7)             (4, 5)
     *      T E X T   B B O X
     *   (0, 1)             (2, 3)     */
    $bbox = imagettfbbox(self::$Data['font-size'], $angle,
                         $this->_getAsset('font'), $text);

    /*
     * example data with angle = 0
     *  [6] => int(-1) [7] => int(-34)     [4] => int(804) [5] => int(-34)
     *  [0] => int(-1) [1] => int(1)       [2] => int(804) [3] => int(1)
     *
     * example data with angle = 90
     *  [6] => int(-34) [7] => int(0)      [4] => int(-34) [5] => int(-806)
     *  [0] => int(1) [1] => int(0)        [2] => int(1) [3] => int(-806)
     */
    switch ($pos_index) {
    case 0: // top text
      $pos_x = (int) (($this->_width - $bbox[4] - $bbox[6]) / 2);
      $pos_y = (int) (-$bbox[7] - self::$Data['text-offset-top']);
      $symbol_x = $pos_x + $bbox[4] - $_symbol_width;
      $symbol_y = $pos_y - (int) (($_symbol_height - $bbox[5]) / 2);
      break;

    case 1: // right text (rotated 90)
      $pos_x = (int) ($this->_width - 1 - $bbox[0]) + self::$Data['text-offset-right'];
      $pos_y = (int) (($this->_height - $bbox[3] + $bbox[1]) / 2);
      $symbol_x = $pos_x - (int) (($_symbol_width - $bbox[4]) / 2);
      $symbol_y = $pos_y + $bbox[5];
      break;

    case 2: // bottom text
      $pos_x = (int) (($this->_width - $bbox[4] - $bbox[6]) / 2);
      $pos_y = (int) ($this->_height - $bbox[1] + self::$Data['text-offset-bottom']);
      $symbol_x = $pos_x + $bbox[4] - $_symbol_width;
      $symbol_y = $pos_y - (int) (($_symbol_height - $bbox[5]) / 2);
      break;

    case 3: // left text (rotated 90)
      $pos_x = (int) (-$bbox[6] - self::$Data['text-offset-left']);
      $pos_y = (int) (($this->_height - $bbox[3] + $bbox[1]) / 2);
      $symbol_x = $pos_x - (int) (($_symbol_width - $bbox[4]) / 2);
      $symbol_y = $pos_y + $bbox[5];
      break;

    default:
      throw new Exception("Invalid position index, must be 0-3");
    }

    imagettftext($this->_image, self::$Data['font-size'], $angle,
                 $pos_x, $pos_y, $this->_textColor, $this->_getAsset('font'),
                 $text);

    /* if there is a symbol to place, figure out where to place it */
    if ($symbol_file) {

      imagecopy($this->_image, $im_symbol,
           $symbol_x, $symbol_y,
           0, 0,
           $_symbol_width, $_symbol_height);
    }
  }

  /**
   * Generates internally an image with the given layout and texts
   *
   * The text is a list with texts in the following order:
   *  - top text
   *  - right text
   *  - bottom text
   *  - left text
   *
   * @param int $layout_index Index of the card layout
   * @param array $texts Text to place on the borders (top, right, bottom, left)
   */
  public function generate($layout_index, $texts) {
    /* clean up any previously generated image */
    if ($this->_image)
      imagedestroy($this->_image);

    /* start from the background */
    $this->_image = imagecreatefrompng($this->_getAsset('background'));
    $this->_width = imagesx($this->_image);
    $this->_height = imagesy($this->_image);
    $this->_textColor = imagecolorallocate($this->_image,
                         self::$Data['text-color'][0],
                         self::$Data['text-color'][1],
                         self::$Data['text-color'][2]);

    /* place the requested layout */
    if (!isset($this->_layouts_resources[$layout_index])) {
      $this->_layout_resources[$layout_index] =
          imagecreatefromjpeg($this->_getAsset('layout-' . $layout_index));
    }
    imagecopy($this->_image, $this->_layout_resources[$layout_index],
        self::$Data['layout-offset-left'], self::$Data['layout-offset-top'],
        0, 0,
        imagesx($this->_layout_resources[$layout_index]),
        imagesy($this->_layout_resources[$layout_index]));

    /* place the texts */
    foreach ($texts as $idx => $text) {
      if ($text != "")
        $this->_writeText($idx, $text);
    }
  }

  /**
   * Save to file or output the generated image
   */
  public function save($filename = null) {
    if ($this->_image === null)
      throw new Exception("No image generated yet!");

    if ($filename === null)
      header("Content-Type: image/jpeg");
    imagejpeg($this->_image, $filename, 90);
  }

}

try {
  $puzzle = new TokenPuzzleRenderer();

  $puzzle->generate(2,
    array("IT IS MY SOLEMN DUTY",
          "AND PLACING THE CROWN UPON HER",
          "THE VEIL. HIS LONG",
          "AS ANY CITIZEN SHOULD. [2]"));
  $puzzle->save();

  // $puzzle->generate(4,
    // array("TO INFORM YOU THAT",
          // "WHOLE CLOTH, FOR WHY WOULD THE",
          // "REIGN HAS COME TO AN",
          // "AND PLACING THE CROWN UPON HER"));
  // $puzzle->save();
}
catch (Exception $e) {
  print "<h1>There is a problem!</h1>\n";
  print "<p>Error: " . htmlspecialchars($e->getMessage()) . "</p>\n";
}
